"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.up = void 0;
async function up(entrances, ...args) {
    let [{ includes, excludes }] = Array.isArray(args[0])
        ? [{ includes: args[0], excludes: [] }]
        : [args[0]];
    let includeMatchers = includes.map(include => buildMatcher(include));
    let excludeMatchers = excludes.map(exclude => buildMatcher(exclude));
    let entryPromises = [];
    for (let key in entrances) {
        if (!includeMatchers.some(matcher => matcher(key)) ||
            excludeMatchers.some(matcher => matcher(key))) {
            continue;
        }
        entryPromises.push(Promise.resolve(entrances[key]).then(value => [key, value]));
    }
    return Object.fromEntries(await Promise.all(entryPromises));
}
exports.up = up;
function buildMatcher(pattern) {
    let starIndex = pattern.indexOf('*');
    if (starIndex < 0) {
        return key => key === pattern;
    }
    else {
        let prefix = pattern.slice(0, starIndex);
        let postfix = pattern.slice(starIndex + 1);
        return key => key.startsWith(prefix) && key.endsWith(postfix);
    }
}
//# sourceMappingURL=utils.js.map